@@ -30,12 +30,12 @@
  * before or not: don't add the tool by hal/udev if user has defined at least
  * one tool for the device in xorg.conf. One device can have multiple tools
  * with the same type to individualize tools with serial number or areas */
-static Bool wcmCheckSource(LocalDevicePtr local, dev_t min_maj)
+static Bool wcmCheckSource(InputInfoPtr pInfo, dev_t min_maj)
 {
 	int match = 0;
 	char* device;
-	char* fsource = xf86CheckStrOption(local->options, "_source", "");
-	LocalDevicePtr pDevices = xf86FirstLocalDevice();
+	char* fsource = xf86CheckStrOption(pInfo->options, "_source", "");
+	InputInfoPtr pDevices = xf86FirstLocalDevice();
 	WacomCommonPtr pCommon = NULL;
 	char* psource;
 
@@ -47,7 +47,7 @@ static Bool wcmCheckSource(LocalDevicePtr local, dev_t min_maj)
 		if (!device || !strstr(pDevices->drv->driverName, "wacom"))
 			continue;
 
-		if (local != pDevices)
+		if (pInfo != pDevices)
 		{
 			psource = xf86CheckStrOption(pDevices->options, "_source", "");
 			pCommon = ((WacomDevicePtr)pDevices->private)->common;
@@ -66,7 +66,7 @@ static Bool wcmCheckSource(LocalDevicePtr local, dev_t min_maj)
 	}
 	if (match)
 		xf86Msg(X_WARNING, "%s: device file already in use by %s. "
-			"Ignoring.\n", local->name, pDevices->name);
+			"Ignoring.\n", pInfo->name, pDevices->name);
 	return match;
 }
 
@@ -78,11 +78,11 @@ static Bool wcmCheckSource(LocalDevicePtr local, dev_t min_maj)
  * the xorg.conf and is then hotplugged through the server backend (HAL,
  * udev). In this case, the hotplugged one fails.
  */
-int wcmIsDuplicate(char* device, LocalDevicePtr local)
+int wcmIsDuplicate(char* device, InputInfoPtr pInfo)
 {
 	struct stat st;
 	int isInUse = 0;
-	char* lsource = xf86CheckStrOption(local->options, "_source", "");
+	char* lsource = xf86CheckStrOption(pInfo->options, "_source", "");
 
 	/* always allow xorg.conf defined tools to be added */
 	if (!strlen(lsource)) goto ret;
@@ -91,7 +91,7 @@ int wcmIsDuplicate(char* device, LocalDevicePtr local)
 	{
 		/* can not access major/minor to check device duplication */
 		xf86Msg(X_ERROR, "%s: stat failed (%s). cannot check for duplicates.\n",
-				local->name, strerror(errno));
+				pInfo->name, strerror(errno));
 
 		/* older systems don't support the required ioctl.  let it pass */
 		goto ret;
@@ -100,7 +100,7 @@ int wcmIsDuplicate(char* device, LocalDevicePtr local)
 	if (st.st_rdev)
 	{
 		/* device matches with another added port */
-		if (wcmCheckSource(local, st.st_rdev))
+		if (wcmCheckSource(pInfo, st.st_rdev))
 		{
 			isInUse = 3;
 			goto ret;
@@ -110,7 +110,7 @@ int wcmIsDuplicate(char* device, LocalDevicePtr local)
 	{
 		/* major/minor can never be 0, right? */
 		xf86Msg(X_ERROR, "%s: device opened with a major/minor of 0. "
-			"Something was wrong.\n", local->name);
+			"Something was wrong.\n", pInfo->name);
 		isInUse = 4;
 	}
 ret:
@@ -120,23 +120,23 @@ ret:
 static struct
 {
 	const char* type;
-	__u16 tool;
+	__u16 tool[3]; /* tool array is terminated by 0 */
 } wcmType [] =
 {
-	{ "stylus", BTN_TOOL_PEN       },
-	{ "eraser", BTN_TOOL_RUBBER    },
-	{ "cursor", BTN_TOOL_MOUSE     },
-	{ "touch",  BTN_TOOL_DOUBLETAP },
-	{ "pad",    BTN_TOOL_FINGER    }
+	{ "stylus", { BTN_TOOL_PEN,       0                  } },
+	{ "eraser", { BTN_TOOL_RUBBER,    0                  } },
+	{ "cursor", { BTN_TOOL_MOUSE,     0                  } },
+	{ "touch",  { BTN_TOOL_DOUBLETAP, BTN_TOOL_FINGER, 0 } },
+	{ "pad",    { BTN_FORWARD,        BTN_0,           0 } }
 };
 
 /* validate tool type for device/product */
-Bool wcmIsAValidType(LocalDevicePtr local, const char* type)
+Bool wcmIsAValidType(InputInfoPtr pInfo, const char* type)
 {
-	int j, ret = FALSE;
-	WacomDevicePtr priv = (WacomDevicePtr)local->private;
+	int j, k, ret = FALSE;
+	WacomDevicePtr priv = (WacomDevicePtr)pInfo->private;
 	WacomCommonPtr common = priv->common;
-	char* dsource = xf86CheckStrOption(local->options, "_source", "");
+	char* dsource = xf86CheckStrOption(pInfo->options, "_source", "");
 
 	if (!type)
 		return FALSE;
@@ -146,17 +146,26 @@ Bool wcmIsAValidType(LocalDevicePtr local, const char* type)
 	{
 		if (!strcmp(wcmType[j].type, type))
 		{
-			if (ISBITSET (common->wcmKeys, wcmType[j].tool))
+			for (k = 0; wcmType[j].tool[k] && !ret; k++)
 			{
-				ret = TRUE;
-				break;
-			}
-			else if (!strlen(dsource)) /* an user defined type */
-			{
-				/* assume it is a valid type */
-				SETBIT(common->wcmKeys, wcmType[j].tool);
-				ret = TRUE;
-				break;
+				if (ISBITSET (common->wcmKeys, wcmType[j].tool[k]))
+				{
+					ret = TRUE;
+
+					/* non GENERIC devices use BTN_TOOL_FINGER for pad */
+					if (common->wcmProtocolLevel != WCM_PROTOCOL_GENERIC)
+					{
+						if (!strcmp(type, "touch") &&
+							wcmType[j].tool[k] == BTN_TOOL_FINGER)
+						    ret = FALSE;
+					}
+				}
+				else if (!strlen(dsource)) /* an user defined type */
+				{
+					/* assume it is a valid type */
+					SETBIT(common->wcmKeys, wcmType[j].tool[k]);
+					ret = TRUE;
+				}
 			}
 		}
 	}
@@ -164,21 +173,26 @@ Bool wcmIsAValidType(LocalDevicePtr local, const char* type)
 }
 
 /* Choose valid types according to device ID. */
-int wcmDeviceTypeKeys(LocalDevicePtr local)
+int wcmDeviceTypeKeys(InputInfoPtr pInfo)
 {
 	int ret = 1;
-	WacomDevicePtr priv = local->private;
+	WacomDevicePtr priv = pInfo->private;
 	WacomCommonPtr common = priv->common;
 
-	priv->common->tablet_id = common->wcmDevCls->ProbeKeys(local);
+	priv->common->tablet_id = common->wcmDevCls->ProbeKeys(pInfo);
 
 	switch (priv->common->tablet_id)
 	{
-		/* tablets with touch ring and rotation pen*/
+		case 0xCC:  /* CintiqV5 */
+			priv->common->tablet_type = WCM_LCD;
+			/* fall through */
+
 		case 0xB8:  /* I4 */
 		case 0xB9:  /* I4 */
 		case 0xBA:  /* I4 */
 		case 0xBB:  /* I4 */
+		case 0xBC:  /* I4 */
+		case 0xBD:  /* I4 */
 			priv->common->tablet_type = WCM_ROTATION;
 			/* fall through */
 
@@ -203,11 +217,30 @@ int wcmDeviceTypeKeys(LocalDevicePtr local)
 			priv->common->tablet_type = WCM_DUALINPUT;
 			break;
 
+		/* P4 display tablets */
+		case 0x30:  /* PL400 */
+		case 0x31:  /* PL500 */
+		case 0x32:  /* PL600 */
+		case 0x33:  /* PL600SX */
+		case 0x34:  /* PL550 */
+		case 0x35:  /* PL800 */
+		case 0x37:  /* PL700 */
+		case 0x38:  /* PL510 */
+		case 0x39:  /* PL710 */
+		case 0xC0:  /* DTF720 */
+		case 0xC2:  /* DTF720a */
+		case 0xC4:  /* DTF521 */
+		case 0xC7:  /* DTU1931 */
+		case 0xCE:  /* DTU2231 */
+		case 0xF0:  /* DTU1631 */
+			priv->common->tablet_type |= WCM_LCD;
+			break;
+
 		/* tablets support menu strips */
 		case 0x3F:  /* CintiqV5 */
 		case 0xC5:  /* CintiqV5 */
 		case 0xC6:  /* CintiqV5 */
-			priv->common->tablet_type |= WCM_LCD;
+			priv->common->tablet_type = WCM_LCD;
 			/* fall through */
 		case 0xB0:  /* I3 */
 		case 0xB1:  /* I3 */
@@ -216,37 +249,56 @@ int wcmDeviceTypeKeys(LocalDevicePtr local)
 		case 0xB4:  /* I3 */
 		case 0xB5:  /* I3 */
 		case 0xB7:  /* I3 */
-			priv->common->tablet_type = WCM_STRIP | WCM_ROTATION;
+			priv->common->tablet_type |= WCM_STRIP | WCM_ROTATION;
 			break;
 
 		case 0xE2: /* TPC with 2FGT */
 		case 0xE3: /* TPC with 2FGT */
+		case 0xE6: /* TPC with 2FGT */
 			priv->common->tablet_type = WCM_TPC;
 			priv->common->tablet_type |= WCM_LCD;
-			/* fall through */
-		case 0xD0:  /* Bamboo with 2FGT */
-		case 0xD1:  /* Bamboo with 2FGT */
-		case 0xD2:  /* Bamboo with 2FGT */
-		case 0xD3:  /* Bamboo with 2FGT */
-			priv->common->tablet_type |= WCM_2FGT;
 			break;
 
 		case 0x93: /* TPC with 1FGT */
 		case 0x9A: /* TPC with 1FGT */
-			priv->common->tablet_type = WCM_1FGT;
-			/* fall through */
 		case 0x90: /* TPC */
 			priv->common->tablet_type |= WCM_TPC;
 			priv->common->tablet_type |= WCM_LCD;
 			break;
 
 		case 0x9F:
-			priv->common->tablet_type = WCM_1FGT;
 			priv->common->tablet_type |= WCM_LCD;
 			break;
+	}
 
-		default:
-			priv->common->tablet_type = WCM_PEN;
+	if (ISBITSET(common->wcmKeys, BTN_TOOL_PEN))
+		priv->common->tablet_type |= WCM_PEN;
+
+	if (ISBITSET (common->wcmKeys, BTN_0) ||
+			ISBITSET (common->wcmKeys, BTN_FORWARD))
+	{
+		priv->common->tablet_type |= WCM_PAD;
+	}
+
+	/* This handles both protocol 4 and 5 meanings of wcmKeys */
+	if (common->wcmProtocolLevel == WCM_PROTOCOL_4)
+	{
+		/* TRIPLETAP means 2 finger touch */
+		/* DOUBLETAP without TRIPLETAP means 1 finger touch */
+		if (ISBITSET(common->wcmKeys, BTN_TOOL_TRIPLETAP))
+			priv->common->tablet_type |= WCM_2FGT;
+		else if (ISBITSET(common->wcmKeys, BTN_TOOL_DOUBLETAP))
+			priv->common->tablet_type |= WCM_1FGT;
+	}
+
+	if (common->wcmProtocolLevel == WCM_PROTOCOL_GENERIC)
+	{
+		/* DOUBLETAP means 2 finger touch */
+		/* FINGER without DOUBLETAP means 1 finger touch */
+		if (ISBITSET(common->wcmKeys, BTN_TOOL_DOUBLETAP))
+			priv->common->tablet_type |= WCM_2FGT;
+		else if (ISBITSET(common->wcmKeys, BTN_TOOL_FINGER))
+			priv->common->tablet_type |= WCM_1FGT;
 	}
 
 	return ret;
@@ -255,31 +307,43 @@ int wcmDeviceTypeKeys(LocalDevicePtr local)
 /**
  * Duplicate xf86 options, replace the "type" option with the given type
  * (and the name with "$name $type" and convert them to InputOption */
-static InputOption *wcmOptionDupConvert(LocalDevicePtr local, const char* basename, const char *type)
+static InputOption *wcmOptionDupConvert(InputInfoPtr pInfo, const char* basename, const char *type)
 {
-	pointer original = local->options;
+	pointer original = pInfo->options;
 	InputOption *iopts = NULL, *new;
-	InputInfoRec dummy;
 	char *name;
+	pointer options;
+	int rc;
 
-	memset(&dummy, 0, sizeof(dummy));
-	xf86CollectInputOptions(&dummy, NULL, original);
+#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 12
+	options = xf86OptionListDuplicate(original);
+#else
+	{
+		InputInfoRec dummy;
 
-	name = Xprintf("%s %s", basename, type);
+		memset(&dummy, 0, sizeof(dummy));
+		xf86CollectInputOptions(&dummy, NULL, original);
+		options = dummy.options;
+	}
+#endif
 
-	dummy.options = xf86ReplaceStrOption(dummy.options, "Type", type);
-	dummy.options = xf86ReplaceStrOption(dummy.options, "Name", name);
+	rc = asprintf(&name, "%s %s", basename, type);
+	if (rc == -1) /* if asprintf fails, strdup will probably too... */
+		name = strdup("unknown");
+
+	options = xf86ReplaceStrOption(options, "Type", type);
+	options = xf86ReplaceStrOption(options, "Name", name);
 	free(name);
 
-	while(dummy.options)
+	while(options)
 	{
 		new = calloc(1, sizeof(InputOption));
 
-		new->key = xf86OptionName(dummy.options);
-		new->value = xf86OptionValue(dummy.options);
+		new->key = xf86OptionName(options);
+		new->value = xf86OptionValue(options);
 		new->next = iopts;
 		iopts = new;
-		dummy.options = xf86NextOption(dummy.options);
+		options = xf86NextOption(options);
 	}
 	return iopts;
 }
@@ -303,68 +367,114 @@ static void wcmFreeInputOpts(InputOption* opts)
  * appended, so a device of product "Wacom" will then have a product "Wacom
  * eraser", "Wacom cursor", etc.
  */
-static InputAttributes* wcmDuplicateAttributes(LocalDevicePtr local,
+static InputAttributes* wcmDuplicateAttributes(InputInfoPtr pInfo,
 					       const char *type)
 {
+	int rc;
 	InputAttributes *attr;
-	attr = DuplicateInputAttributes(local->attrs);
-	attr->product = Xprintf("%s %s", attr->product, type);
+	attr = DuplicateInputAttributes(pInfo->attrs);
+	rc = asprintf(&attr->product, "%s %s", attr->product, type);
+	if (rc == -1)
+		attr->product = NULL;
 	return attr;
 }
 #endif
 
 /**
- * Hotplug one device of the given type.
+ * This struct contains the necessary info for hotplugging a device later.
+ * Memory must be freed after use.
+ */
+typedef struct {
+	InputOption *input_options;
+#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 9
+	InputAttributes *attrs;
+#endif
+} WacomHotplugInfo;
+
+/**
+ * Actually hotplug the device. This function is called by the server when
+ * the WorkProcs are processed.
+ *
+ * @param client The server client. unused
+ * @param closure A pointer to a struct WcmHotplugInfo containing the
+ * necessary information to create a new device.
+ * @return TRUE to remove this function from the server's work queue.
+ */
+static Bool
+wcmHotplugDevice(ClientPtr client, pointer closure )
+{
+	WacomHotplugInfo *hotplug_info = closure;
+	DeviceIntPtr dev; /* dummy */
+
+	NewInputDeviceRequest(hotplug_info->input_options,
+#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 9
+			      hotplug_info->attrs,
+#endif
+			      &dev);
+	wcmFreeInputOpts(hotplug_info->input_options);
+
+#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 11
+	FreeInputAttributes(hotplug_info->attrs);
+#endif
+	free(hotplug_info);
+
+	return TRUE;
+}
+
+/**
+ * Queue the hotplug for one tool/device of the given type.
  * Device has the same options as the "parent" device, type is one of
  * erasor, stylus, pad, touch, cursor, etc.
  * Name of the new device is set automatically to "<device name> <type>".
+ *
+ * Note that we don't actually hotplug the device here. We store the
+ * information needed to hotplug the device later and then queue the
+ * hotplug. The server will come back and call the @wcmHotplugDevice
+ * later.
+ *
+ * @param pInfo The parent device
+ * @param basename The base name for the device (type will be appended)
+ * @param type Type name for this tool
  */
-static void wcmHotplug(LocalDevicePtr local, const char* basename, const char *type)
+static void wcmQueueHotplug(InputInfoPtr pInfo, const char* basename, const char *type)
 {
-	DeviceIntPtr dev; /* dummy */
-	InputOption *input_options;
-#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 9
-	InputAttributes *attrs = NULL;
-#endif
+	WacomHotplugInfo *hotplug_info;
 
-	input_options = wcmOptionDupConvert(local, basename, type);
+	hotplug_info = calloc(1, sizeof(WacomHotplugInfo));
 
+	if (!hotplug_info)
+	{
+		xf86Msg(X_ERROR, "%s: OOM, cannot hotplug dependent devices\n", pInfo->name);
+		return;
+	}
+
+	hotplug_info->input_options = wcmOptionDupConvert(pInfo, basename, type);
 #if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 11
-	attrs = wcmDuplicateAttributes(local, type);
-#endif
-
-	NewInputDeviceRequest(input_options,
-#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 9
-				attrs,
-#endif
-				&dev);
-	wcmFreeInputOpts(input_options);
-
-#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 11
-	FreeInputAttributes(attrs);
+	hotplug_info->attrs = wcmDuplicateAttributes(pInfo, type);
 #endif
+	QueueWorkProc(wcmHotplugDevice, serverClient, hotplug_info);
 }
 
-void wcmHotplugOthers(LocalDevicePtr local, const char *basename)
+void wcmHotplugOthers(InputInfoPtr pInfo, const char *basename)
 {
 	int i, skip = 1;
 	char*		device;
 
-        xf86Msg(X_INFO, "%s: hotplugging dependent devices.\n", local->name);
-	device = xf86SetStrOption(local->options, "Device", NULL);
+        xf86Msg(X_INFO, "%s: hotplugging dependent devices.\n", pInfo->name);
+	device = xf86SetStrOption(pInfo->options, "Device", NULL);
         /* same loop is used to init the first device, if we get here we
          * need to start at the second one */
 	for (i = 0; i < ARRAY_SIZE(wcmType); i++)
 	{
-		if (wcmIsAValidType(local, wcmType[i].type))
+		if (wcmIsAValidType(pInfo, wcmType[i].type))
 		{
 			if (skip)
 				skip = 0;
 			else
-				wcmHotplug(local, basename, wcmType[i].type);
+				wcmQueueHotplug(pInfo, basename, wcmType[i].type);
 		}
 	}
-        xf86Msg(X_INFO, "%s: hotplugging completed.\n", local->name);
+        xf86Msg(X_INFO, "%s: hotplugging completed.\n", pInfo->name);
 }
 
 /**
@@ -376,9 +486,9 @@ void wcmHotplugOthers(LocalDevicePtr local, const char *basename)
  * This changes the source to _driver/wacom, all auto-hotplugged devices
  * will have the same source.
  */
-int wcmNeedAutoHotplug(LocalDevicePtr local, const char **type)
+int wcmNeedAutoHotplug(InputInfoPtr pInfo, const char **type)
 {
-	char *source = xf86CheckStrOption(local->options, "_source", "");
+	char *source = xf86CheckStrOption(pInfo->options, "_source", "");
 	int i;
 
 	if (*type) /* type specified, don't hotplug */
@@ -391,7 +501,7 @@ int wcmNeedAutoHotplug(LocalDevicePtr local, const char **type)
 	 * for our device */
 	for (i = 0; i < ARRAY_SIZE(wcmType); i++)
 	{
-		if (wcmIsAValidType(local, wcmType[i].type))
+		if (wcmIsAValidType(pInfo, wcmType[i].type))
 		{
 			*type = strdup(wcmType[i].type);
 			break;
@@ -401,47 +511,59 @@ int wcmNeedAutoHotplug(LocalDevicePtr local, const char **type)
 	if (!*type)
 		return 0;
 
-	xf86Msg(X_INFO, "%s: type not specified, assuming '%s'.\n", local->name, *type);
-	xf86Msg(X_INFO, "%s: other types will be automatically added.\n", local->name);
+	xf86Msg(X_INFO, "%s: type not specified, assuming '%s'.\n", pInfo->name, *type);
+	xf86Msg(X_INFO, "%s: other types will be automatically added.\n", pInfo->name);
 
-	local->options = xf86AddNewOption(local->options, "Type", *type);
-	local->options = xf86ReplaceStrOption(local->options, "_source", "_driver/wacom");
+	/* Note: wcmIsHotpluggedDevice() relies on this */
+	pInfo->options = xf86AddNewOption(pInfo->options, "Type", *type);
+	pInfo->options = xf86ReplaceStrOption(pInfo->options, "_source", "_driver/wacom");
 
 	return 1;
 }
 
-int wcmParseOptions(LocalDevicePtr local, int hotplugged)
+/**
+ * Parse the options for this device.
+ * Note that parameters is_primary and is_dependent are mutually exclusive,
+ * though both may be false in the case of an xorg.conf device.
+ *
+ * @param is_primary True if the device is the parent device for
+ * hotplugging, False if the device is a depent or xorg.conf device.
+ * @param is_hotplugged True if the device is a dependent device, FALSE
+ * otherwise.
+ * @retvalue True on success or False otherwise.
+ */
+Bool wcmParseOptions(InputInfoPtr pInfo, Bool is_primary, Bool is_dependent)
 {
-	WacomDevicePtr  priv = (WacomDevicePtr)local->private;
+	WacomDevicePtr  priv = (WacomDevicePtr)pInfo->private;
 	WacomCommonPtr  common = priv->common;
 	char            *s, b[12];
-	int		i, oldButton;
+	int		i;
 	WacomToolPtr    tool = NULL;
-	WacomToolAreaPtr area = NULL;
+	int		tpc_button_is_on;
 
 	/* Optional configuration */
-	priv->debugLevel = xf86SetIntOption(local->options,
+	priv->debugLevel = xf86SetIntOption(pInfo->options,
 					    "DebugLevel", priv->debugLevel);
-	common->debugLevel = xf86SetIntOption(local->options,
+	common->debugLevel = xf86SetIntOption(pInfo->options,
 					      "CommonDBG", common->debugLevel);
-	s = xf86SetStrOption(local->options, "Mode", NULL);
+	s = xf86SetStrOption(pInfo->options, "Mode", NULL);
 
 	if (s && (xf86NameCmp(s, "absolute") == 0))
-		set_absolute(local, TRUE);
+		set_absolute(pInfo, TRUE);
 	else if (s && (xf86NameCmp(s, "relative") == 0))
-		set_absolute(local, FALSE);
+		set_absolute(pInfo, FALSE);
 	else
 	{
 		if (s)
 			xf86Msg(X_ERROR, "%s: invalid Mode (should be absolute"
-				" or relative). Using default.\n", local->name);
+				" or relative). Using default.\n", pInfo->name);
 
 		/* If Mode not specified or is invalid then rely on
 		 * Type specific defaults from initialization.
 		 */
 	}
 
-	/* Pad is always in relative mode.
+	/* Pad is always in absolute mode.
 	 * The pad also defaults to wheel scrolling, unlike the pens
 	 * (interesting effects happen on ArtPen and others with build-in
 	 * wheels)
@@ -450,32 +572,45 @@ int wcmParseOptions(LocalDevicePtr local, int hotplugged)
 	{
 		priv->wheelup = 4;
 		priv->wheeldn = 5;
-		set_absolute(local, FALSE);
+		set_absolute(pInfo, TRUE);
 	}
 
-	/* Store original local Core flag so it can be changed later */
-	if (local->flags & (XI86_ALWAYS_CORE | XI86_CORE_POINTER))
-		priv->flags |= COREEVENT_FLAG;
-
-	s = xf86SetStrOption(local->options, "Rotate", NULL);
+	s = xf86SetStrOption(pInfo->options, "Rotate", NULL);
 
 	if (s)
 	{
+		int rotation;
+
 		if (xf86NameCmp(s, "CW") == 0)
-			common->wcmRotate=ROTATE_CW;
+			rotation = ROTATE_CW;
 		else if (xf86NameCmp(s, "CCW") ==0)
-			common->wcmRotate=ROTATE_CCW;
+			rotation = ROTATE_CCW;
 		else if (xf86NameCmp(s, "HALF") ==0)
-			common->wcmRotate=ROTATE_HALF;
+			rotation = ROTATE_HALF;
 		else if (xf86NameCmp(s, "NONE") !=0)
 		{
 			xf86Msg(X_ERROR, "%s: invalid Rotate option '%s'.\n",
-				local->name, s);
+				pInfo->name, s);
 			goto error;
 		}
+
+		if (is_dependent && rotation != common->wcmRotate)
+			xf86Msg(X_INFO, "%s: ignoring rotation of dependent"
+					" device\n", pInfo->name);
+		else
+			wcmRotateTablet(pInfo, rotation);
 	}
 
-	common->wcmSuppress = xf86SetIntOption(local->options, "Suppress",
+	common->wcmRawSample = xf86SetIntOption(pInfo->options, "RawSample",
+			common->wcmRawSample);
+	if (common->wcmRawSample < 1 || common->wcmRawSample > MAX_SAMPLES)
+	{
+		xf86Msg(X_ERROR, "%s: RawSample setting '%d' out of range [1..%d]. Using default.\n",
+			pInfo->name, common->wcmRawSample, MAX_SAMPLES);
+		common->wcmRawSample = DEFAULT_SAMPLES;
+	}
+
+	common->wcmSuppress = xf86SetIntOption(pInfo->options, "Suppress",
 			common->wcmSuppress);
 	if (common->wcmSuppress != 0) /* 0 disables suppression */
 	{
@@ -485,79 +620,44 @@ int wcmParseOptions(LocalDevicePtr local, int hotplugged)
 			common->wcmSuppress = DEFAULT_SUPPRESS;
 	}
 
-	if (xf86SetBoolOption(local->options, "Tilt",
+	if (xf86SetBoolOption(pInfo->options, "Tilt",
 			(common->wcmFlags & TILT_REQUEST_FLAG)))
 		common->wcmFlags |= TILT_REQUEST_FLAG;
 
-	if (xf86SetBoolOption(local->options, "RawFilter",
-			(common->wcmFlags & RAW_FILTERING_FLAG)))
-		common->wcmFlags |= RAW_FILTERING_FLAG;
-
 	/* pressure curve takes control points x1,y1,x2,y2
 	 * values in range from 0..100.
 	 * Linear curve is 0,0,100,100
 	 * Slightly depressed curve might be 5,0,100,95
 	 * Slightly raised curve might be 0,5,95,100
 	 */
-	s = xf86SetStrOption(local->options, "PressCurve", "0,0,100,100");
+	s = xf86SetStrOption(pInfo->options, "PressCurve", "0,0,100,100");
 	if (s && (IsStylus(priv) || IsEraser(priv)))
 	{
 		int a,b,c,d;
 		if ((sscanf(s,"%d,%d,%d,%d",&a,&b,&c,&d) != 4) ||
 				!wcmCheckPressureCurveValues(a, b, c, d))
 			xf86Msg(X_CONFIG, "%s: PressCurve not valid\n",
-				local->name);
+				pInfo->name);
 		else
 			wcmSetPressureCurve(priv,a,b,c,d);
 	}
 
 	if (IsCursor(priv))
 	{
-		common->wcmCursorProxoutDist = xf86SetIntOption(local->options, "CursorProx", 0);
-		if (common->wcmCursorProxoutDist < 0 || common->wcmCursorProxoutDist > 255)
+		common->wcmCursorProxoutDist = xf86SetIntOption(pInfo->options, "CursorProx", 0);
+		if (common->wcmCursorProxoutDist < 0 ||
+				common->wcmCursorProxoutDist > common->wcmMaxDist)
 			xf86Msg(X_CONFIG, "%s: CursorProx invalid %d \n",
-				local->name, common->wcmCursorProxoutDist);
+				pInfo->name, common->wcmCursorProxoutDist);
 	}
 
-	/* Configure Monitors' resoluiton in TwinView setup.
-	 * The value is in the form of "1024x768,1280x1024"
-	 * for a desktop of monitor 1 at 1024x768 and
-	 * monitor 2 at 1280x1024
-	 */
-	s = xf86SetStrOption(local->options, "TVResolution", NULL);
-	if (s)
-	{
-		int a,b,c,d;
-		if ((sscanf(s,"%dx%d,%dx%d",&a,&b,&c,&d) != 4) ||
-			(a <= 0) || (b <= 0) || (c <= 0) || (d <= 0))
-			xf86Msg(X_CONFIG, "%s: TVResolution not valid\n",
-				local->name);
-		else
-		{
-			priv->tvResolution[0] = a;
-			priv->tvResolution[1] = b;
-			priv->tvResolution[2] = c;
-			priv->tvResolution[3] = d;
-		}
-	}
-
-	priv->screen_no = xf86SetIntOption(local->options, "ScreenNo", -1);
-
-	if (xf86SetBoolOption(local->options, "KeepShape", 0))
-		priv->flags |= KEEP_SHAPE_FLAG;
-
-	priv->topX = xf86SetIntOption(local->options, "TopX", 0);
-	priv->topY = xf86SetIntOption(local->options, "TopY", 0);
-	priv->bottomX = xf86SetIntOption(local->options, "BottomX", 0);
-	priv->bottomY = xf86SetIntOption(local->options, "BottomY", 0);
-	priv->serial = xf86SetIntOption(local->options, "Serial", 0);
+	priv->topX = xf86SetIntOption(pInfo->options, "TopX", 0);
+	priv->topY = xf86SetIntOption(pInfo->options, "TopY", 0);
+	priv->bottomX = xf86SetIntOption(pInfo->options, "BottomX", 0);
+	priv->bottomY = xf86SetIntOption(pInfo->options, "BottomY", 0);
+	priv->serial = xf86SetIntOption(pInfo->options, "Serial", 0);
 
 	tool = priv->tool;
-	area = priv->toolarea;
-	area->topX = priv->topX;
-	area->topY = priv->topY;
-	area->bottomX = priv->bottomX;
-	area->bottomY = priv->bottomY;
 	tool->serial = priv->serial;
 
 	/* The first device doesn't need to add any tools/areas as it
@@ -573,18 +673,10 @@ int wcmParseOptions(LocalDevicePtr local, int hotplugged)
 
 		if(toollist) /* Already have a tool with the same type/serial */
 		{
-			WacomToolAreaPtr arealist;
-
-			free(tool);
-			priv->tool = tool = toollist;
-			arealist = toollist->arealist;
-
-			/* Add the area to the end of the list */
-			while(arealist->next)
-				arealist = arealist->next;
-			arealist->next = area;
-		}
-		else /* No match on existing tool/serial, add tool to the end of the list */
+			xf86Msg(X_ERROR, "%s: already have a tool with type/serial %d/%d.",
+					pInfo->name, tool->typeid, tool->serial);
+			goto error;
+		} else /* No match on existing tool/serial, add tool to the end of the list */
 		{
 			toollist = common->wcmTool;
 			while(toollist->next)
@@ -593,94 +685,84 @@ int wcmParseOptions(LocalDevicePtr local, int hotplugged)
 		}
 	}
 
-	common->wcmThreshold = xf86SetIntOption(local->options, "Threshold",
+	common->wcmThreshold = xf86SetIntOption(pInfo->options, "Threshold",
 			common->wcmThreshold);
 
-	common->wcmMaxZ = xf86SetIntOption(local->options, "MaxZ",
+	common->wcmMaxZ = xf86SetIntOption(pInfo->options, "MaxZ",
 					   common->wcmMaxZ);
-	if (xf86SetBoolOption(local->options, "ButtonsOnly", 0))
+	if (xf86SetBoolOption(pInfo->options, "ButtonsOnly", 0))
 		priv->flags |= BUTTONS_ONLY_FLAG;
 
 	/* TPCButton on for Tablet PC by default */
-	if (TabletHasFeature(common, WCM_TPC))
-		common->wcmTPCButtonDefault = 1;
+	tpc_button_is_on = xf86SetBoolOption(pInfo->options, "TPCButton",
+					TabletHasFeature(common, WCM_TPC));
 
-	oldButton = xf86SetBoolOption(local->options, "TPCButton",
-					common->wcmTPCButtonDefault);
-
-	if (hotplugged || IsStylus(priv))
-		common->wcmTPCButton = oldButton;
-	else if (oldButton != common->wcmTPCButton)
+	if (is_primary || IsStylus(priv))
+		common->wcmTPCButton = tpc_button_is_on;
+	else if (tpc_button_is_on != common->wcmTPCButton)
 		xf86Msg(X_WARNING, "%s: TPCButton option can only be set "
-			"by stylus.\n", local->name);
+			"by stylus.\n", pInfo->name);
 
-	/* a single touch device */
-	if (ISBITSET (common->wcmKeys, BTN_TOOL_DOUBLETAP))
+	/* a single or double touch device */
+	if (TabletHasFeature(common, WCM_1FGT) ||
+	    TabletHasFeature(common, WCM_2FGT))
 	{
+		int touch_is_on, capacity_is_on;
+
 		/* TouchDefault was off for all devices
 		 * except when touch is supported */
 		common->wcmTouchDefault = 1;
 
-		oldButton = xf86SetBoolOption(local->options, "Touch",
-					common->wcmTouchDefault);
+		touch_is_on = xf86SetBoolOption(pInfo->options, "Touch",
+						common->wcmTouchDefault);
 
-		if (hotplugged || IsTouch(priv))
-			common->wcmTouch = oldButton;
-		else if (oldButton != common->wcmTouch)
+		if (is_primary || IsTouch(priv))
+			common->wcmTouch = touch_is_on;
+		else if (touch_is_on != common->wcmTouch)
 			xf86Msg(X_WARNING, "%s: Touch option can only be set "
-				"by a touch tool.\n", local->name);
+				"by a touch tool.\n", pInfo->name);
 
-		oldButton = xf86SetBoolOption(local->options, "Capacity",
-					common->wcmCapacityDefault);
+		capacity_is_on = xf86SetBoolOption(pInfo->options, "Capacity",
+						   common->wcmCapacityDefault);
 
-		if (hotplugged || IsTouch(priv))
-			common->wcmCapacity = oldButton;
-		else if (oldButton != common->wcmCapacity)
+		if (is_primary || IsTouch(priv))
+			common->wcmCapacity = capacity_is_on;
+		else if (capacity_is_on != common->wcmCapacity)
 			xf86Msg(X_WARNING, "%s: Touch Capacity option can only be"
-				"set by a touch tool.\n", local->name);
+				"set by a touch tool.\n", pInfo->name);
 	}
 
 	/* 2FG touch device */
-	if (ISBITSET (common->wcmKeys, BTN_TOOL_TRIPLETAP))
+	if (TabletHasFeature(common, WCM_2FGT))
 	{
+		int gesture_is_on;
+
 		/* GestureDefault was off for all devices
 		 * except when multi-touch is supported */
 		common->wcmGestureDefault = 1;
 
-		oldButton = xf86SetBoolOption(local->options, "Gesture",
-					common->wcmGestureDefault);
+		gesture_is_on = xf86SetBoolOption(pInfo->options, "Gesture",
+					    common->wcmGestureDefault);
 
-		if (hotplugged || IsTouch(priv))
-			common->wcmGesture = oldButton;
-		else if (oldButton != common->wcmGesture)
+		if (is_primary || IsTouch(priv))
+			common->wcmGesture = gesture_is_on;
+		else if (gesture_is_on != common->wcmGesture)
 			xf86Msg(X_WARNING, "%s: Touch gesture option can only "
-				"be set by a touch tool.\n", local->name);
-
-		if ((common->wcmDevCls == &gWacomUSBDevice) &&
-				TabletHasFeature(common, WCM_LCD) &&
-				TabletHasFeature(common, WCM_2FGT)) {
-			common->wcmGestureParameters.wcmZoomDistanceDefault = 30;
-			common->wcmGestureParameters.wcmScrollDistanceDefault = 30;
-			common->wcmGestureParameters.wcmTapTimeDefault = 250;
-		}
+				"be set by a touch tool.\n", pInfo->name);
 
 		common->wcmGestureParameters.wcmZoomDistance =
-			xf86SetIntOption(local->options, "ZoomDistance",
+			xf86SetIntOption(pInfo->options, "ZoomDistance",
 			common->wcmGestureParameters.wcmZoomDistanceDefault);
 
 		common->wcmGestureParameters.wcmScrollDistance =
-			xf86SetIntOption(local->options, "ScrollDistance",
+			xf86SetIntOption(pInfo->options, "ScrollDistance",
 			common->wcmGestureParameters.wcmScrollDistanceDefault);
 
 		common->wcmGestureParameters.wcmTapTime =
-			xf86SetIntOption(local->options, "TapTime",
+			xf86SetIntOption(pInfo->options, "TapTime",
 			common->wcmGestureParameters.wcmTapTimeDefault);
 	}
 
-	/* Mouse cursor stays in one monitor in a multimonitor setup */
-	if ( !priv->wcmMMonitor )
-		priv->wcmMMonitor = xf86SetBoolOption(local->options, "MMonitor", 1);
-
 	/* Swap stylus buttons 2 and 3 for Tablet PCs */
 	if (TabletHasFeature(common, WCM_TPC) && IsStylus(priv))
 	{
@@ -691,45 +773,18 @@ int wcmParseOptions(LocalDevicePtr local, int hotplugged)
 	for (i=0; i<WCM_MAX_BUTTONS; i++)
 	{
 		sprintf(b, "Button%d", i+1);
-		priv->button[i] = xf86SetIntOption(local->options, b, priv->button[i]);
+		priv->button[i] = xf86SetIntOption(pInfo->options, b, priv->button[i]);
 	}
 
-	s = xf86SetStrOption(local->options, "Twinview", NULL);
-	if (s && xf86NameCmp(s, "none") == 0)
-		priv->twinview = TV_NONE;
-	else if ((s && xf86NameCmp(s, "horizontal") == 0) ||
-			(s && xf86NameCmp(s, "rightof") == 0))
-		priv->twinview = TV_LEFT_RIGHT;
-	else if ((s && xf86NameCmp(s, "vertical") == 0) ||
-			(s && xf86NameCmp(s, "belowof") == 0))
-		priv->twinview = TV_ABOVE_BELOW;
-	else if (s && xf86NameCmp(s, "leftof") == 0)
-		priv->twinview = TV_RIGHT_LEFT;
-	else if (s && xf86NameCmp(s, "aboveof") == 0)
-		priv->twinview = TV_BELOW_ABOVE;
-	else if (s)
-	{
-		xf86Msg(X_ERROR, "%s: invalid Twinview (should be none, vertical (belowof), "
-			"horizontal (rightof), aboveof, or leftof). Using none.\n",
-			local->name);
-		priv->twinview = TV_NONE;
-	}
-
-	if (s && priv->twinview != TV_NONE)
-		priv->numScreen = 2;
-	else
-		priv->numScreen = screenInfo.numScreens;
-
 	/* Now parse class-specific options */
 	if (common->wcmDevCls->ParseOptions &&
-	    !common->wcmDevCls->ParseOptions(local))
+	    !common->wcmDevCls->ParseOptions(pInfo))
 		goto error;
 
-	return 1;
+	return TRUE;
 error:
-	free(area);
 	free(tool);
-	return 0;
+	return FALSE;
 }
 
-/* vim: set noexpandtab shiftwidth=8: */
+/* vim: set noexpandtab tabstop=8 shiftwidth=8: */
